@@ -1,11 +1,17 @@
 # -*- coding: utf-8 -*-
-import re
+from __future__ import unicode_literals
+
 import os
+
+
 extensions = [
+    'sphinx.ext.autodoc',
+    'sphinx.ext.autosummary',
     'sphinx.ext.todo',
     'sphinx.ext.coverage',
     'sphinx.ext.ifconfig',
     'sphinx.ext.viewcode',
+    'sphinxcontrib.napoleon'
 ]
 if os.getenv('SPELLCHECK'):
     extensions += 'sphinxcontrib.spelling',
@@ -14,16 +20,17 @@ if os.getenv('SPELLCHECK'):
 
 source_suffix = '.rst'
 master_doc = 'index'
-project = u'lazy-object-proxy'
-copyright = u'2014, Ionel Cristian Mărieș'
-version = release = re.findall(
-    'version="(.*)"',
-    open(os.path.join(os.path.dirname(__file__), '../setup.py')).read()
-)[0]
-
+project = 'lazy-object-proxy'
+year = '2014-2015'
+author = 'Ionel Cristian Mărieș'
+copyright = '{0}, {1}'.format(year, author)
+version = release = '1.2.1'
 import sphinx_py3doc_enhanced_theme
 html_theme = "sphinx_py3doc_enhanced_theme"
 html_theme_path = [sphinx_py3doc_enhanced_theme.get_html_theme_path()]
+html_theme_options = {
+    'githuburl': 'https://github.com/ionelmc/python-lazy-object-proxy/'
+}
 
 pygments_style = 'trac'
 templates_path = ['.']
@@ -34,6 +41,3 @@ html_sidebars = {
    '**': ['searchbox.html', 'globaltoc.html', 'sourcelink.html'],
 }
 html_short_title = '%s-%s' % (project, version)
-html_theme_options = {
-    'githuburl': 'https://github.com/ionelmc/python-lazy-object-proxy/'
-}
